/*
Copyright (C) 2008 Jeremy Chadwick. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/param.h>
#include <dev/smbus/smb.h>
#include "global.h"


/* Function prototypes */
unsigned int read_byte(int, int, const char);
unsigned int w83792d_divisor(unsigned int n);
unsigned int w83792d_rpmconv(unsigned int, unsigned int);
void w83792d_read(int, int, struct sensors *);
unsigned int w83793g_rpmconv(unsigned int);
unsigned int w83793g_tempadj(unsigned int);
void w83793g_read(int, int, struct sensors *);
void x6dva_read(int, struct sensors *);
void data_collect(int, struct board *, struct sensors *);

/* External functions (main.c) */
extern void VERBOSE(const char *, ...);
extern void DIE(const char *fmt, ...);

/* Global variables */
static char ibuf[SMB_MAXBLOCKSIZE];	/* Buffer for SMBus data reads */
static unsigned char regmap[256];	/* Array for registers/indexes */


/*
 * read_byte(int fd, int slave, const char mycmd)
 *
 *    fd = Descriptor return from open() on a /dev/smbX device
 * slave = SMBus slave address; see boardlist[] in boards.c
 * mycmd = Index/offset to read; see collect.c for hos this is used
 * 
 * Reads a byte off off the SMBus via ioctl().  See smb(4) for details.
 *
 * Returns byte read.  On failure, returns -1.
 */
unsigned int read_byte(int fd, int slave, const char mycmd)
{
	struct smbcmd c;

	c.slave = slave;
	c.cmd = mycmd;
	c.data.byte_ptr = ibuf;

	if (ioctl(fd, SMB_READB, &c) == -1) {
		return -1;
	}

	return (int)(unsigned char) ibuf[0];
}


/*
 * w83792d_divisor(unsigned int n)
 *
 * n = PWM input divisor
 *
 * Converts a Winbond W83792D fan PWM input divisor to an integer value.
 * Supposedly the default value is 2, so we assume that in the case the
 * divisor is > 7.
 *
 * Returns a numeric divisor, which can be used mathematically.
 */
unsigned int w83792d_divisor(unsigned int n)
{
        VERBOSE("entering w83792d_divisor(0x%02x)\n", n);

	switch (n) {
		case 0x00: return 1;
		case 0x01: return 2;
		case 0x02: return 4;
		case 0x03: return 8;
		case 0x04: return 16;
		case 0x05: return 32;
		case 0x06: return 64;
		case 0x07: return 128;
	}
	return 2;
}


/*
 * w83792d_rpmconv(unsigned int count, unsigned int div)
 *
 * count = Winbond register: fan speed counter
 *   div = Winbond register: fan divisor
 *
 * Converts Winbond W83792D fan speed counter register to RPMs.  The
 * conversion formula, according to Winbond, is:
 *
 *   RPM = (1.35 * (10^6)) / (count * divisor)
 *
 * A count value of 0xff (255) means the fan is disconnected.  We also
 * make sure count and div are non-zero, ensuring we don't divide by zero.
 *
 * Returns the current revolutions-per-minute (RPM) of the fan.  If
 * the fan is disconnected, or either count or div are zero, return 0.
 */
unsigned int w83792d_rpmconv(unsigned int count, unsigned int div)
{
        VERBOSE("entering w83792d_rpmconv(0x%02x, %u)\n", count, div);

	if (count != 0xff && count != 0 && div != 0) {
		return 1350000 / (count * div);
	}
	return 0;
}


/*
 * w83792d_read(int fd, int slave, struct sensors *s)
 *
 *    fd = Descriptor return from open() on a /dev/smbX device
 * slave = SMBus slave address; see boardlist[] in boards.c
 *     s = Pointer to sensors struct; see global.h for a definition
 *
 * Winbond W83792D register reading subroutine.  This does the bulk of
 * the work.  Any board which uses the W83792D will use this routine to
 * read a series of registers (called "CRxx") off the SMBus.
 *
 * The registers we're interested in are scattered all over the place,
 * which makes writing a clean-looking function very difficult.  There
 * are probably better ways to do this (such as contain a map of all
 * the registers).  At least they're all in Bank 0!  Here's the list:
 *
 *   CR20-CR2A
 *   CR3E-CR3F
 *   CR47
 *   CR58-CR5C
 *   CRB0-CRB1
 *   CRC0
 *   CRC8
 *
 * For details of what register serves what purpose, refer to the
 * official Winbond W83792D documentation (April 26, 2006; rev 0.9).
 * I've included a decent "map" of what register does what in the
 * below comments, for quick reference/debugging.
 */
void w83792d_read(int fd, int slave, struct sensors *s)
{
	unsigned int i;
	unsigned int fandiv;

        VERBOSE("entering w83792d_read(%d, 0x%02x)\n", fd, slave);

	for (i = 0x20; i <= 0x2a; i++) {
		regmap[i] = read_byte(fd, slave, i);
		VERBOSE("CR%02X value: 0x%02x\n", i, regmap[i]);
	}

	regmap[0x3e] = read_byte(fd, slave, 0x3e);
	regmap[0x3f] = read_byte(fd, slave, 0x3f);
	regmap[0x47] = read_byte(fd, slave, 0x47);
	VERBOSE("CR3E value: 0x%02x\n", regmap[0x3e]);
	VERBOSE("CR3F value: 0x%02x\n", regmap[0x3f]);
	VERBOSE("CR47 value: 0x%02x\n", regmap[0x47]);

	for (i = 0x58; i <= 0x5c; i++) {
		regmap[i] = read_byte(fd, slave, i);
		VERBOSE("CR%02X value: 0x%02x\n", i, regmap[i]);
	}

	regmap[0xb0] = read_byte(fd, slave, 0xb0);
	regmap[0xb1] = read_byte(fd, slave, 0xb1);
	regmap[0xc0] = read_byte(fd, slave, 0xc0);
	regmap[0xc8] = read_byte(fd, slave, 0xc8);
	VERBOSE("CRB0 value: 0x%02x\n", regmap[0xb0]);
	VERBOSE("CRB1 value: 0x%02x\n", regmap[0xb1]);
	VERBOSE("CRC0 value: 0x%02x\n", regmap[0xc0]);
	VERBOSE("CRC8 value: 0x%02x\n", regmap[0xc8]);

	/*
	 * Winbond pin    Indexes used
	 * ------------   ---------------------
	 * VCOREA         0x20, 0x3e (bits 1-0)
	 * VCOREB         0x21, 0x3e (bits 3-2)
	 * VIN0           0x22, 0x3e (bits 5-4)
	 * VIN1           0x23, 0x3e (bits 7-6)
	 * VIN2           0x24, 0x3f (bits 1-0)
	 * VIN3           0x25, 0x3f (bits 3-2)
	 * 5VCC           0x26, 0x3f (bits 5-4) XXX - Calculation formula wrong?
	 * 5VSB           0xb0
	 * VBAT           0xb1
	 * ------------   ---------------------
	 *
	 * The bits used for CR3E (VCOREB) are undocumented; I assume 3-2.
	 *
	 * The calculation formulas used below are suspect.  The formulas given in
	 * the W83792D data sheet are either incorrect, or Supermicro chose to use
	 * adifferent resistors than what Winbond did.
	 *
	 * The problematic one appears to be VIN1, which on my systems shows up as
	 * 3.040V or so, which definitely is wrong.
	 *
	 * Additionally, the documentation Supermicro provided does not state
	 * anything about 5VSB or VBAT, but they're shown in the BIOS, and testing
	 * shows they're connected.
	 */
	s->voltages[VOLT_VCOREA].value = ((regmap[0x20] << 2) + (regmap[0x3e] & 0x03)) * 0.002;
	s->voltages[VOLT_VCOREB].value = ((regmap[0x21] << 2) + ((regmap[0x3e] & 0x0c) >> 2)) * 0.002;
	s->voltages[VOLT_VIN0].value = ((regmap[0x22] << 2) + ((regmap[0x3e] & 0x30) >> 4)) * 0.004;
	s->voltages[VOLT_VIN1].value = ((regmap[0x23] << 2) + ((regmap[0x3e] & 0xc0) >> 6)) * 0.004;
	s->voltages[VOLT_VIN2].value = ((regmap[0x24] << 2) + (regmap[0x3f] & 0x03)) * 0.008 * 12;
	s->voltages[VOLT_VIN2].value *= -1;		/* VIN2 is a negative voltage */
	s->voltages[VOLT_VIN3].value = ((regmap[0x25] << 2) + ((regmap[0x3f] & 0x0c) >> 2)) * 0.016;
	s->voltages[VOLT_5VCC].value = ((regmap[0x26] << 2) + ((regmap[0x3f] & 0x30) >> 4)) * 0.006;
	s->voltages[VOLT_5VSB].value = regmap[0xb0] * 0.024;
	s->voltages[VOLT_VBAT].value = regmap[0xb1] * 0.016;

	/*
	 * Winbond pin    Indexes used
	 * ------------   ---------------------
	 * TD1            0x27
	 * TD2            0xc0, 0xc1 (bit 7)
	 * TD3            0xc8, 0xc9 (bit 7)
	 * ------------   ---------------------
	 *
	 * The official Winbond W83792D documentation says some of the temperature sensors
	 * are 9-bit, where the last bit defines .5 or .0.  However, Supermicro didn't
	 * bother to follow any of this, and instead use and instead made all the values
	 * raw 8-bit.  This is why we don't read CRC1 or CRC9.
	 */
	s->temps[TEMP_TD1].value = regmap[0x27];
	s->temps[TEMP_TD2].value = regmap[0xc0];
	s->temps[TEMP_TD3].value = regmap[0xc8];

	/*
	 * Winbond pin    Indexes used
	 * ------------   ---------------------
	 * FAN1           0x28  (divisor at 0x47, bits 2-0)
	 * FAN2           0x29  (divisor at 0x47, bits 6-4)
	 * FAN3           0x2a  (divisor at 0x5b, bits 2-0)
	 * FAN4           0xb8  (divisor at 0x5b, bits 6-4)
	 * FAN5           0xb9  (divisor at 0x5c, bits 2-0)
	 * FAN6           0xba  (divisor at 0x5c, bits 6-4)
	 * FAN7           0xbe  (divisor at 0x9e, bits 2-0)
	 * ------------   ---------------------
	 */ 
	fandiv = w83792d_divisor(regmap[0x47] & 0x07);
	s->fans[FAN_FAN1].value = w83792d_rpmconv(regmap[0x28], fandiv);

	fandiv = w83792d_divisor((regmap[0x47] & 0x70) >> 4);
	s->fans[FAN_FAN2].value = w83792d_rpmconv(regmap[0x29], fandiv);

	fandiv = w83792d_divisor(regmap[0x5b] & 0x07);
	s->fans[FAN_FAN3].value = w83792d_rpmconv(regmap[0x2a], fandiv);

	fandiv = w83792d_divisor((regmap[0x5b] & 0x70) >> 4);
	s->fans[FAN_FAN4].value = w83792d_rpmconv(regmap[0xb8], fandiv);

	fandiv = w83792d_divisor(regmap[0x5c] & 0x07);
	s->fans[FAN_FAN5].value = w83792d_rpmconv(regmap[0xb9], fandiv);

	fandiv = w83792d_divisor((regmap[0x5c] & 0x70) >> 4);
	s->fans[FAN_FAN6].value = w83792d_rpmconv(regmap[0xba], fandiv);

	fandiv = w83792d_divisor(regmap[0x9e] & 0x07);
	s->fans[FAN_FAN7].value = w83792d_rpmconv(regmap[0xbe], fandiv);

	VERBOSE("exiting w83792d_read()\n");
}


/*
 * w83793g_rpmconv(unsigned int count)
 *
 * count = Winbond register: fan speed counter
 *
 * Converts Winbond W83793G fan speed counter register to RPMs.  The
 * conversion formula, according to Winbond, is:
 *
 *   RPM = (1.35 * (10^6)) / (count * (fanpoles/4))
 *
 * "fanpoles" is undocumented, and is completely dependent upon the
 * actual fan model used.  Most fan motors are 4-pole, although some
 * (very) old models could be 2-pole.  "Pole" refers to actual polarity,
 * e.g. north or south.  We have to assume all fans connected are 4-pole
 * models.  There is no * way to auto-detect pole count.  For details
 * on how fan motors work and what a "pole" is, see the below URL (which
 * also describes why 4-pole motors are better than 2 or 3-pole):
 *
 * http://electronics.howstuffworks.com/motor6.htm
 *
 * A count value of 0x0fff (4095) means the fan is disconnected.  We also
 * make sure count is non-zero, ensuring we don't divide by zero.
 *
 * Returns the current revolutions-per-minute (RPM) of the fan.  If
 * the fan is disconnected, or count is 0, return 0.
 */
unsigned int w83793g_rpmconv(unsigned int count)
{
        VERBOSE("entering w83793g_rpmconv(0x%04x)\n", count);

	if (count != 0x0fff && count != 0) {
		return 1350000 / count;
	}
	return 0;
}


/*
 * w83793g_tempadj(unsigned int raw)
 *
 * raw = Winbond register: temperature register
 *
 * Winbond W83793G temperatures have 10 bits of granularity: 1 sign bit
 * (defining position or negative temperature), 7 data bits (for the
 * integer portion of the temperature), and 2 remainder bits (for the
 * floating point portion).
 *
 * bsdhwmon makes two assumptions, to keep things simple:
 *
 * 1) We only bother to read the registers used for the integer portion,
 * and the signage.  We're not interested in the floating point part of
 * the temperature.  This allows us to avoid switching banks and other
 * complexities, sticking with a pure "read-only" model,
 *
 * 2) The sign bit/MSB should NEVER be set (a temperature inside of a PC
 * should never be so cold as to have a negative temperature; a processor
 * should not be -72C, for example).  If the MSB is set, bsdhwmon assumes
 * there's no SMBus register tie-ins, or there's no thermistor installed.
 *
 * With regards to #2, Supermicro boards appear to set the MSB when there
 * is no temperature being monitored -- so we rely on that.  The author
 * has personally confirmed this.
 *
 * Returns the value passed to the function, but only if the MSB isn't
 * set.  If the MSB is set, return 0.
 */
unsigned int w83793g_tempadj(unsigned int raw)
{
	if ((raw & 0x80) == 0x80) {
		return 0;
	}
	return raw;
}


/*
 * w83793g_read(int fd, int slave, struct sensors *s)
 *
 *    fd = Descriptor return from open() on a /dev/smbX device
 * slave = SMBus slave address; see boardlist[] in boards.c
 *     s = Pointer to sensors struct; see global.h for a definition
 *
 * Winbond W83793G register reading subroutine.  This does the bulk of
 * the work.  Any board which uses the W83793G will use this routine to
 * read a series of registers (called "CRxx") off the SMBus.
 *
 * The registers we're interested in are CR10 through CR3A in bank 0.
 *
 * For details of what register serves what purpose, refer to the
 * official Winbond W83793G documentation (December 11, 2006; rev 1.0).
 * I've included a decent "map" of what register does what in the
 * below comments, for quick reference/debugging.
 */
void w83793g_read(int fd, int slave, struct sensors *s)
{
	unsigned int i;

        VERBOSE("entering w83793g_read(%d, 0x%02x)\n", fd, slave);

	for (i = 0x10; i <= 0x3a; i++) {
		regmap[i] = read_byte(fd, slave, i);
		VERBOSE("CR%02X value: 0x%02x\n", i, regmap[i]);
	}

	/*
	 * Winbond pin    Registers used
	 * ------------   ---------------------
	 * VCOREA         0x10, 0x1b (bits 1-0)
	 * VCOREB         0x11, 0x1b (bits 3-2)
	 * VTT            0x12, 0x1b (bits 5-4)
	 * VSEN1          0x14
	 * VSEN2          0x15
	 * 3VSEN          0x16
	 * 12VSEN         0x17
	 * 5VDD           0x18
	 * 5VSB           0x19
	 * VBAT           0x1a
	 * ------------   ---------------------
	 *
	 * The calculation formulas used below are suspect.  The formulas given in
	 * the W83793G data sheet are either incorrect, or Supermicro chose to use
	 * adifferent resistors than what Winbond did.
	 *
	 * Thanks to Jim Perry for helping with some of the calculations.
	 */
	s->voltages[VOLT_VCOREA].value = ((regmap[0x10] << 2) + (regmap[0x1b] & 0x03)) * 0.002;
	s->voltages[VOLT_VCOREB].value = ((regmap[0x11] << 2) + ((regmap[0x1b] & 0x0c) >> 2)) * 0.002;
	s->voltages[VOLT_VTT].value    = ((regmap[0x12] << 2) + ((regmap[0x1b] & 0x30) >> 4)) * 0.002;
	s->voltages[VOLT_VSEN1].value  = regmap[0x14] * 0.032 * 12;
	s->voltages[VOLT_VSEN1].value  *= -1;		/* VSEN1 is a negative voltage */
	s->voltages[VOLT_VSEN2].value  = regmap[0x15] * 0.016;
	s->voltages[VOLT_3VSEN].value  = regmap[0x16] * 0.016;
	s->voltages[VOLT_12VSEN].value = regmap[0x17] * 0.008 * 12;
	s->voltages[VOLT_5VDD].value   = (regmap[0x18] * 0.024) + 0.15;
	s->voltages[VOLT_5VSB].value   = (regmap[0x19] * 0.024) + 0.15;
	s->voltages[VOLT_VBAT].value   = regmap[0x1a] * 0.016;

	/*
	 * Winbond pin    Registers used
	 * ------------   ---------------------
	 * TD1            0x1c, 0x22 (bits 1-0)
	 * TD2            0x1d, 0x22 (bits 3-2)
	 * TD3            0x1e, 0x22 (bits 5-4)
	 * TD4            0x1f, 0x22 (bits 7-6)
	 * TR1            0x20
	 * TR2            0x21
	 * ------------   ---------------------
	 *
	 * TD temperatures are 10 bits: 1 sign bit (MSB), 7 data bits, and 2 decimal
	 * bits.  We're not interested in the decimal portion, but we are interested
	 * in the integer portion.  tempadj() checks the MSB (sign bit) and if it's
	 * set, makes the assumption that there's no wire/tie-in.
	 *
	 * TR temperatures are 1 sign bit (MSB), 7 data bits.
	 */
	s->temps[TEMP_TD1].value = w83793g_tempadj(regmap[0x1c]);
	s->temps[TEMP_TD2].value = w83793g_tempadj(regmap[0x1d]);
	s->temps[TEMP_TD3].value = w83793g_tempadj(regmap[0x1e]);
	s->temps[TEMP_TD4].value = w83793g_tempadj(regmap[0x1f]);
	s->temps[TEMP_TR1].value = regmap[0x20];
	s->temps[TEMP_TR2].value = regmap[0x21];

	/* See the official W83793G specification sheet for these. */
	s->fans[FAN_FAN1].value = w83793g_rpmconv((regmap[0x23] << 8) | regmap[0x24]);
	s->fans[FAN_FAN2].value = w83793g_rpmconv((regmap[0x25] << 8) | regmap[0x26]);
	s->fans[FAN_FAN3].value = w83793g_rpmconv((regmap[0x27] << 8) | regmap[0x28]);
	s->fans[FAN_FAN4].value = w83793g_rpmconv((regmap[0x29] << 8) | regmap[0x2a]);
	s->fans[FAN_FAN5].value = w83793g_rpmconv((regmap[0x2b] << 8) | regmap[0x2c]);
	s->fans[FAN_FAN6].value = w83793g_rpmconv((regmap[0x2d] << 8) | regmap[0x2e]);
	s->fans[FAN_FAN7].value = w83793g_rpmconv((regmap[0x2f] << 8) | regmap[0x30]);
	s->fans[FAN_FAN8].value = w83793g_rpmconv((regmap[0x31] << 8) | regmap[0x32]);
	s->fans[FAN_FAN9].value = w83793g_rpmconv((regmap[0x33] << 8) | regmap[0x34]);
	s->fans[FAN_FAN10].value = w83793g_rpmconv((regmap[0x35] << 8) | regmap[0x36]);
	s->fans[FAN_FAN11].value = w83793g_rpmconv((regmap[0x37] << 8) | regmap[0x38]);
	s->fans[FAN_FAN12].value = w83793g_rpmconv((regmap[0x39] << 8) | regmap[0x3a]);

        VERBOSE("exiting w83793g_read()\n");
}


/*
 * x6dva_read(int fd, struct sensors *)
 *
 * fd = Descriptor return from open() on a /dev/smbX device
 *  s = Pointer to sensors struct; see global.h for a definition
 *
 * Supermicro X6DVA / X6DVL / X6DAL register reading subroutine.
 *
 * These Supermicro boards are quite custom/unique -- they use dual
 * Winbond chips (one W83627HF, and one W83792D).  This makes things very
 * complex when it comes to reading IC data.
 *
 * We do not want to risk using the standard w83792d_read() routine, since
 * that may read some registers which aren't tied to anything and could
 * put system stability at risk.
 *
 * The SMBus slave addresses are 0x2c for the W83627HF, and 0x2f for the
 * W83792D.  These are assumed/hard-coded in the below function, rather
 * than passed in via boardlist[].
 *
 * The registers we care about are:
 *
 *   Winbond W83627HF registers:
 *     CR20-CR25, CR27
 *   Winbond W83792D registers:
 *     CR20-CR21, CR28-CR2A, CR3E, CR47, CR5B-5C, CRB8-CRBA, CRC0, CRC8
 */
void x6dva_read(int fd, struct sensors *s)
{
	int w83627hf_slave = 0x2c;
	int w83792d_slave = 0x2f;
	unsigned int fandiv;

        VERBOSE("entering x6dva_read(%d)\n", fd);

	/* Winbond W83627HF portion */
	/*
	 * Winbond pin    Indexes used   Purpose   Variable used
	 * ------------   ------------   -------   -------------
	 * VCOREA         0x20           +1.5V     VOLT_VIN0
	 * VCOREB         0x21           +3.3VSB   VOLT_VIN1
	 * +3.3VIN        0x22           +3.3V     VOLT_VIN2
	 * +5VIN          0x23           +5V       VOLT_VIN3
	 * +12VIN         0x24           +12V      VOLT_12VSEN
	 * -12VIN         0x25           -12V      VOLT_VSEN1
	 * unknown        0x27           SysTemp   TEMP_TD1
	 * ------------   ------------   -------   -------------
	 */
	regmap[0x20] = read_byte(fd, w83627hf_slave, 0x20);
	regmap[0x21] = read_byte(fd, w83627hf_slave, 0x21);
	regmap[0x22] = read_byte(fd, w83627hf_slave, 0x22);
	regmap[0x23] = read_byte(fd, w83627hf_slave, 0x23);
	regmap[0x24] = read_byte(fd, w83627hf_slave, 0x24);
	regmap[0x25] = read_byte(fd, w83627hf_slave, 0x25);
	regmap[0x25] = read_byte(fd, w83627hf_slave, 0x25);
	regmap[0x27] = read_byte(fd, w83627hf_slave, 0x27);
	VERBOSE("CR20 value: 0x%02x\n", regmap[0x20]);
	VERBOSE("CR21 value: 0x%02x\n", regmap[0x21]);
	VERBOSE("CR22 value: 0x%02x\n", regmap[0x22]);
	VERBOSE("CR23 value: 0x%02x\n", regmap[0x23]);
	VERBOSE("CR24 value: 0x%02x\n", regmap[0x24]);
	VERBOSE("CR25 value: 0x%02x\n", regmap[0x25]);
	VERBOSE("CR27 value: 0x%02x\n", regmap[0x27]);

	/* We now need to calculate voltages and temperatures which use the
	 * populated regmap[] array.  Do this BEFORE the Winbond W83792D
	 * reads, since some regmap indexes are re-used (e.g. 0x20).
	 */
	s->voltages[VOLT_VIN0].value   = regmap[0x20];
	s->voltages[VOLT_VIN1].value   = regmap[0x21];
	s->voltages[VOLT_VIN2].value   = regmap[0x22];
	s->voltages[VOLT_VIN3].value   = regmap[0x23];
	s->voltages[VOLT_12VSEN].value = regmap[0x24];
	s->voltages[VOLT_VSEN1].value  = regmap[0x25];
	s->voltages[VOLT_VSEN1].value  *= -1;		/* VSEN1 is a negative voltage */
	s->temps[TEMP_TD1].value = regmap[0x27];


	/* Winbond W83792D portion */

	regmap[0x20] = read_byte(fd, w83792d_slave, 0x20);
	regmap[0x21] = read_byte(fd, w83792d_slave, 0x21);
	regmap[0x28] = read_byte(fd, w83792d_slave, 0x28);
	regmap[0x29] = read_byte(fd, w83792d_slave, 0x29);
	regmap[0x2a] = read_byte(fd, w83792d_slave, 0x2a);
	regmap[0x3e] = read_byte(fd, w83792d_slave, 0x3e);
	regmap[0x47] = read_byte(fd, w83792d_slave, 0x47);
	regmap[0x5b] = read_byte(fd, w83792d_slave, 0x5b);
	regmap[0x5c] = read_byte(fd, w83792d_slave, 0x5c);
	regmap[0xb8] = read_byte(fd, w83792d_slave, 0xb8);
	regmap[0xb9] = read_byte(fd, w83792d_slave, 0xb9);
	regmap[0xba] = read_byte(fd, w83792d_slave, 0xba);
	regmap[0xc0] = read_byte(fd, w83792d_slave, 0xc0);
	regmap[0xc8] = read_byte(fd, w83792d_slave, 0xc8);
	VERBOSE("CR20 value: 0x%02x\n", regmap[0x20]);
	VERBOSE("CR21 value: 0x%02x\n", regmap[0x21]);
	VERBOSE("CR28 value: 0x%02x\n", regmap[0x28]);
	VERBOSE("CR29 value: 0x%02x\n", regmap[0x29]);
	VERBOSE("CR2A value: 0x%02x\n", regmap[0x2a]);
	VERBOSE("CR3E value: 0x%02x\n", regmap[0x3e]);
	VERBOSE("CR47 value: 0x%02x\n", regmap[0x47]);
	VERBOSE("CR5B value: 0x%02x\n", regmap[0x5b]);
	VERBOSE("CR5C value: 0x%02x\n", regmap[0x5c]);
	VERBOSE("CRB8 value: 0x%02x\n", regmap[0xb8]);
	VERBOSE("CRB9 value: 0x%02x\n", regmap[0xb9]);
	VERBOSE("CRBA value: 0x%02x\n", regmap[0xba]);
	VERBOSE("CRC0 value: 0x%02x\n", regmap[0xc0]);
	VERBOSE("CRC8 value: 0x%02x\n", regmap[0xc8]);

	/*
	 * Winbond pin    Indexes used
	 * ------------   ---------------------
	 * VCOREA         0x20, 0x3e (bits 1-0)
	 * VCOREB         0x21, 0x3e (bits 3-2)
	 * ------------   ---------------------
	 *
	 * The bits used for CR3E (VCOREB) are undocumented; I assume 3-2.
	 *
	 * The calculation formulas used below are suspect.  The formulas given in
	 * the W83792D data sheet are either incorrect, or Supermicro chose to use
	 * adifferent resistors than what Winbond did.
	 */
	s->voltages[VOLT_VCOREA].value = ((regmap[0x20] << 2) + (regmap[0x3e] & 0x03)) * 0.002;
	s->voltages[VOLT_VCOREB].value = ((regmap[0x21] << 2) + ((regmap[0x3e] & 0x0c) >> 2)) * 0.002;

	/*
	 * Winbond pin    Indexes used
	 * ------------   ---------------------
	 * TD2            0xc0, 0xc1 (bit 7)
	 * TD3            0xc8, 0xc9 (bit 7)
	 * ------------   ---------------------
	 *
	 * The official Winbond W83792D documentation says some of the temperature sensors
	 * are 9-bit, where the last bit defines .5 or .0.  However, Supermicro didn't
	 * bother to follow any of this, and instead use and instead made all the values
	 * raw 8-bit.  This is why we don't read CRC1 or CRC9.
	 */
	s->temps[TEMP_TD2].value = regmap[0xc0];
	s->temps[TEMP_TD3].value = regmap[0xc8];

	/*
	 * Winbond pin    Indexes used
	 * ------------   ---------------------
	 * FAN1           0x28  (divisor at 0x47, bits 2-0)
	 * FAN2           0x29  (divisor at 0x47, bits 6-4)
	 * FAN3           0x2a  (divisor at 0x5b, bits 2-0)
	 * FAN4           0xb8  (divisor at 0x5b, bits 6-4)
	 * FAN5           0xb9  (divisor at 0x5c, bits 2-0)
	 * FAN6           0xba  (divisor at 0x5c, bits 6-4)
	 * ------------   ---------------------
	 */ 
	fandiv = w83792d_divisor(regmap[0x47] & 0x07);
	s->fans[FAN_FAN1].value = w83792d_rpmconv(regmap[0x28], fandiv);

	fandiv = w83792d_divisor((regmap[0x47] & 0x70) >> 4);
	s->fans[FAN_FAN2].value = w83792d_rpmconv(regmap[0x29], fandiv);

	fandiv = w83792d_divisor(regmap[0x5b] & 0x07);
	s->fans[FAN_FAN3].value = w83792d_rpmconv(regmap[0x2a], fandiv);

	fandiv = w83792d_divisor((regmap[0x5b] & 0x70) >> 4);
	s->fans[FAN_FAN4].value = w83792d_rpmconv(regmap[0xb8], fandiv);

	fandiv = w83792d_divisor(regmap[0x5c] & 0x07);
	s->fans[FAN_FAN5].value = w83792d_rpmconv(regmap[0xb9], fandiv);

	fandiv = w83792d_divisor((regmap[0x5c] & 0x70) >> 4);
	s->fans[FAN_FAN6].value = w83792d_rpmconv(regmap[0xba], fandiv);

        VERBOSE("exiting x6dva_read()\n");
}


/*
 * data_collect(int fd, struct board *b, struct sensors *s)
 *
 * fd = Descriptor return from open() on a /dev/smbX device
 *  b = Pointers to board struct; see boards.c for a definition
 *  s = Pointer to sensors struct; see global.h for a definition
 *
 * This routine is what's called by main() once a board model has been
 * matched.  We examine the "chip" variable of the board structure, and
 * compare it against a list of known chip types (see global.h, enum
 * chips_e).
 *
 * A match should ALWAYS be found here -- otherwise, something very bad
 * is happening upstream in the program (in main()), or a user has
 * tinkered with the code without knowing what they're doing.
 */
void data_collect(int fd, struct board *b, struct sensors *s)
{
        VERBOSE("entering data_collect(%d)\n", fd);

	switch (b->chip) {
		case CUSTOM_X6DVA:
			x6dva_read(fd, s);
			break;
		case WINBOND_W83792D:
			w83792d_read(fd, b->slave, s);
			break;
		case WINBOND_W83793G:
			w83793g_read(fd, b->slave, s);
			break;
		default:
			/* XXX - Should never get here */
			DIE("hit 'default' case() in data_collect()");
			break;
	}
        VERBOSE("exiting data_collect()\n");
}

