/*
Copyright (C) 2008 Jeremy Chadwick. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <sys/param.h>
#include <kenv.h>
#include "global.h"


#define DEFAULT_SMBDEV		"/dev/smb0"

/* Function prototypes */
void VERBOSE(const char *, ...);
void DIE(const char *, ...);
void USAGE(void);

/* External functions (boards.c) */
extern struct board *board_lookup(const char *);

/* External functions (collect.c) */
extern void data_collect(int, struct board *, struct sensors *);

/* External functions (output.c) */
extern void sensors_output(struct board *, struct sensors *);
extern void list_models(struct board *);

/* External variables (boards.c) */
extern struct board boardlist;

/* Global variables */
int smbfd;					/* File descriptor for /dev/smbXXX */
char smbdev[MAXPATHLEN] = DEFAULT_SMBDEV;	/* Command line flag "-f", otherwise /dev/smb0 */
int f_verbose = 0;				/* Command line flag "-v" */


/*
 * VERBOSE() is mainly used for debugging.  Use of the "-v" flag will
 * cause underlying functions to call VERBOSE() with all sorts of
 * information.
 */
void VERBOSE(const char *fmt, ...)
{
	if (f_verbose) {
		va_list ap;

		va_start(ap, fmt);
		printf("==> ");
		vprintf(fmt, ap);
		va_end(ap);
	}
}


/*
 * DIE() is used in cases where an error has occurred.  An exit code
 * of 1 is always assumed, whereas zero means success.
 */
void DIE(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	printf("ERROR: ");
	vprintf(fmt, ap);
	va_end(ap);
	exit(1);
}


void USAGE(void)
{
	printf("Usage: bsdhwmon [options]\n");
	printf("\n");
	printf("Options:\n");
	printf("\n");
	printf("  -f DEVICE     use DEVICE as smb(4) device (default: " DEFAULT_SMBDEV ")\n");
	printf("  -l            list supported motherboard ID strings\n");
	printf("  -h            print this message\n");
	printf("  -v            be verbose\n");
	printf("\n");
	printf("bsdhwmon version %u\n", VERSION);
	printf("http://bsdhwmon.parodius.com/\n");
	printf("Report bugs to <koitsu@FreeBSD.org>\n");
	exit(0);
}


int main(int argc, char *argv[])
{
	int ch;
	char kenv_planar_product[] = "smbios.planar.product";
	char *boardname;
	struct board *mb;
	struct sensors *sdata;

	if ((boardname = calloc(1, MAXPATHLEN)) == NULL) {
		DIE("calloc() for boardname failed: %s\n",
			strerror(errno)
		);
	}

	while ((ch = getopt(argc, argv, "f:lvh?")) != -1) {
		switch (ch) {
			case 'f':
				strncpy(smbdev, optarg, MAXPATHLEN);
				break;
			case 'l':
				list_models(&boardlist);
				break;
			case 'v':
				f_verbose = 1;
				break;
			case 'h':
			case '?':
			default:
				USAGE();
		}
	}

	argc -= optind;
	argv += optind;

	/*
	 * bsdhwmon requires root access, since it has to open a /dev/smbX device.
	 */
	if (geteuid() != 0) {
		DIE("Must be run as root, or setuid root.\n");
	}

	/*
	 * Use kenv(2) to look up the smbios.planar.product string and return it.
	 * This is the motherboard string provided by the BIOS.  If SMBIOS information
	 * isn't available to the user, then they're out of luck.
	 */
	if (kenv(KENV_GET, kenv_planar_product, boardname, MAXPATHLEN) == -1) {
		DIE("kenv() lookup for %s failed: %s\n",
			kenv_planar_product,
			strerror(errno)
		);
	}

	if ((mb = board_lookup(boardname)) == NULL) {
		DIE("Board '%s' is not presently supported.\n\n"
		    "You can visit http://bsdhwmon.parodius.com/ to see if support\n"
		    "for your motherboard/system is under development.\n",
			boardname
		);
	}

	if ((sdata = calloc(1, sizeof(struct sensors))) == NULL) {
		DIE("calloc() for sdata failed: %s\n",
			strerror(errno)
		);
	}
	VERBOSE("Sensors struct at %p\n", sdata);

	if ((smbfd = open(smbdev, O_RDWR)) < 0) {
		DIE("open() %s failed: %s\n",
			smbdev,
			strerror(errno)
		);
	}

	/* Collect all of the data via SMBus registers. */
	data_collect(smbfd, mb, sdata);

	/*
	 * We always assume success on the close().  I have seen situations (in
	 * other unrelated software) where sometimes close() will fail due to
	 * odd reasons.  Rather than emit an error, just continue on silently.
	 */
	close(smbfd);

	/* Now output the collected data to the user. */
	sensors_output(mb, sdata);

	/* Clean up and exit. */
	free(sdata);
	exit(0);
}

